import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;

/**
 * A small application do demonstrate the use of the 
 * FileHistory class.
 * 
 * @since JDK1.3 
 * @author Klaus Berg
 */
public class DemoFileHistory extends JFrame
    implements FileHistory.IFileHistory {

    private static final String FILE_SEPARATOR_STR = System.getProperty("file.separator");
    private JMenuBar menuBar;
    private JToolBar toolBar;
    private JMenu fileMenu, optionsMenu;
    private FileHistory fileHistory;
    private JLabel status;
    private TextCanvas canvas;

    /**
     * Constructor.
     */
    public DemoFileHistory() {
        super("Demo File History");
        // change look&feel to Cross platfor L&F
        try {
            UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
        fileMenu = new JMenu("File");
        fileMenu.setMnemonic('F');
        toolBar = new JToolBar();
        menuBar = new JMenuBar();
        status = new JLabel("Status line");
        status.setBorder(BorderFactory.createLoweredBevelBorder());
        this.setJMenuBar(menuBar);
        menuBar.add(fileMenu);

        NewAction newAction = new NewAction();
        fileMenu.add(newAction);

        OpenAction openAction = new OpenAction();
        fileMenu.add(openAction);

        JMenuItem exitMenuItem = new JMenuItem("Exit");
        exitMenuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                exit();
            }
        });
        KeyStroke ksExit = KeyStroke.getKeyStroke(KeyEvent.VK_X, Event.CTRL_MASK);
        exitMenuItem.setAccelerator(ksExit);
        fileMenu.add(exitMenuItem);
        menuBar.add(fileMenu);

        optionsMenu = new JMenu("Options");
        optionsMenu.setMnemonic('O');
        HistoryManagementAction historyManagementAction = new HistoryManagementAction();
        optionsMenu.add(historyManagementAction);
        menuBar.add(optionsMenu);

        JButton newB = new JButton();
        newB.setAction(newAction);
        newB.setIcon(new ImageIcon(DemoFileHistory.class.getResource("new.gif"))); // must be called AFTER setAction()!     
        toolBar.add(newB);
        JButton openB = new JButton();      
        openB.setAction(openAction);
        openB.setIcon(new ImageIcon(DemoFileHistory.class.getResource("open.gif"))); // must be called AFTER setAction()!
        toolBar.add(openB);

        fileHistory = new FileHistory(this); // init FileHistory
        fileHistory.initFileMenuHistory();

        this.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                exit();
            }
        });
        this.getContentPane().add(toolBar, BorderLayout.NORTH);
        canvas = new TextCanvas();
        this.getContentPane().add(canvas);
        this.getContentPane().add(status, BorderLayout.SOUTH);
        this.pack();
        this.setVisible(true);
    }

    // --- Implementation of FileHistory.IFileHistory interface ----------------

    /**
     * Get the application name to identify the configuration file in the
     * the USER_HOME directory. This name should be unique in this directory.
     * 
     * @return the application name
     */
    public String getApplicationName() {
        return "DEMO";
    }

    /**
     * Get a handle to the frame's file menu.
     * 
     * @return the frame's file menu
     */
    public JMenu getFileMenu() {
        return menuBar.getMenu(0);
    }

    /**
     * Return the size of the main application frame.
     * It is used to center the file history maintenance window.
     * 
     * @return the frame's size
     */
    public Dimension getSize() {
        return super.getSize();
    }

    /**
     * Return the main application frame.
     * It is used to center the file history maintenance window.
     * 
     * @return the main GUI frame
     */
    public JFrame getFrame() {
        return this;
    }

    /**
     * Simulate a load file activity.
     * 
     * @param path   the pathname of the loaded file
     */
    public void loadFile(String path) {
        status.setText("File loaded: "+path);
        String filename = path.substring(path.lastIndexOf(FILE_SEPARATOR_STR)+1,
                              path.length());
        if (filename.length() > 30) {
            filename = filename.substring(0, 30) + "...";
        }
        canvas.updateText(filename);
        canvas.repaint();
    }

    // -------------------------------------------------------------------------

    /**
     * Close the frame, save the file history entries and exit the demo.
     */
    private void exit() {
        setVisible(false);
        dispose();
        fileHistory.saveHistoryEntries(); // save entries for next session
        System.exit(0);
    }

    /**
     * The application's main method to test it.
     * 
     * @param argv   - not used -
     */
    public static void main(String[] argv) {
        new DemoFileHistory();
    }


    // --- Action classes ------------------------------------------------------
    private final class NewAction extends AbstractAction {

        NewAction() {
            super("New...");
        }

        public void actionPerformed(ActionEvent e) {
            status.setText("New file selected");
            canvas.updateText("");
            canvas.repaint();
        }
    }

    private final class OpenAction extends AbstractAction {

        OpenAction() {
            super("Open...");
        }

        public void actionPerformed(ActionEvent e) {
            JFileChooser fileChooser = new JFileChooser();
            fileChooser.setCurrentDirectory(new File(System.getProperty("user.dir")));
            int result = fileChooser.showOpenDialog(DemoFileHistory.this);
            if (result == JFileChooser.APPROVE_OPTION) {
                File f = fileChooser.getSelectedFile();
                if (f == null) {
                    return;
                }
                if (f.getName().equals("")) {
                    return;
                }
                String path = f.getAbsolutePath();
                status.setText("Open " + path);            
                fileHistory.insertPathname(path); // hook into FileHistory class
                loadFile(path);
            }
        }
    }

    private final class HistoryManagementAction extends AbstractAction {

        HistoryManagementAction() {
            super("File History...");                  
        }

        public void actionPerformed(ActionEvent e) {
            fileHistory.processList(); // hook into FileHistory class
        }
    }

    /**
     * A helper class to render the loaded filename in the center
     * of the frame. We use some methods from the Java 2D API to
     * create a mirror effect.
     * It's just to simulate an activity performed with the loaded
     * filename that is provided with the updateText() method.
     */
    private final class TextCanvas extends JPanel {

        String text = "File History Demo Application";
        Font font = new Font("serif", Font.BOLD | Font.ITALIC, 36);

        TextCanvas() {
            this.setBackground(Color.white);
            setPreferredSize(new Dimension(600, 350));
        }

        public void updateText(String text) {
            this.text = text;
        }      

        public void paintComponent(Graphics _g) {
            Graphics2D g = (Graphics2D)_g;
            g.clearRect(0, 0, this.getWidth(), this.getHeight());
            AffineTransform mirror = AffineTransform.getScaleInstance(1, -1);
            Font mirroredFont = font.deriveFont(mirror);
            g.setFont(font);
            g.setPaint(Color.blue);
            g.drawString(text, 20, 170);
            g.setFont(mirroredFont);
            g.setPaint(new Color(0, 0, 0, 128)); // 50% transparent black
            g.drawString(text, 20, 170);
        }
    }

}

